(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RootLabeling imports SemanticLabeling Name begin

definition root_altering :: "('f,'v)trs \<Rightarrow> ('f,'v)trs" ("'(_')a" 1000)
where "(R)a \<equiv> {(l,r) | l r. (l,r) \<in> R \<and> root l \<noteq> root r}"

text {* The following function creates a flat context with root @{text f}
and a hole at position @{text i}. All variables are taken from the list
@{text xs}. *}
definition flat_ctxt_i :: "('f,string)term list \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> ('f,string)ctxt"
where "flat_ctxt_i xs f i \<equiv> More f (take i xs) \<box> (drop (Suc i) xs)"

definition fresh_vars :: "nat \<Rightarrow> string set \<Rightarrow> string list"
where "fresh_vars n tabus \<equiv> fresh_strings ''x'' 0 tabus n"

definition fresh_vars_list :: "nat \<Rightarrow> string list \<Rightarrow> string list"
where "fresh_vars_list n tabus \<equiv> fresh_strings_list ''x'' 0 tabus n"

lemma fresh_vars_list_sound[simp]:
  "fresh_vars_list n tabus = fresh_vars n (set tabus)"
  unfolding fresh_vars_list_def fresh_vars_def by auto

lemma flat_ctxt_i_fresh_vars_list_conv[simp]:
  "flat_ctxt_i(map Var (fresh_vars_list n tabus)) = flat_ctxt_i(map Var (fresh_vars n (set tabus)))"
by (rule ext,rule ext,unfold flat_ctxt_i_def) simp

lemma fresh_vars_fresh:
  assumes "finite V"
  shows "set(fresh_vars n V) \<inter> V = {}"
  using assms fresh_name_gen_for_strings[of "''x''" 0]
  unfolding fresh_vars_def fresh_name_gen_def by auto

lemma fresh_vars_list_fresh: "set(fresh_vars_list n vs) \<inter> set vs = {}"
  using fresh_name_gen_for_strings_list[of "''x''" 0]
  unfolding fresh_vars_list_def fresh_name_gen_list_def by auto

lemma fresh_vars_length:
  assumes "finite V" shows "length(fresh_vars n V) = n"
  using assms fresh_name_gen_for_strings[of "''x''" 0]
  unfolding fresh_name_gen_def fresh_vars_def by simp

lemma fresh_vars_list_length: "length(fresh_vars_list n vs) = n"
  using fresh_name_gen_for_strings_list[of "''x''" 0]
  unfolding fresh_name_gen_list_def fresh_vars_list_def by simp

lemma Var_inj: "inj_on Var (set vs)" unfolding inj_on_def by auto

lemma fresh_vars_distinct: assumes "finite V" shows "distinct(fresh_vars n V)"
  using assms Var_inj[of "fresh_strings ''x'' 0 V n"]
    fresh_name_gen_for_strings[of "''x''" 0]
    distinct_map[of Var "fresh_strings ''x'' 0 V n"]
  unfolding fresh_name_gen_def fresh_vars_def by auto

lemma fresh_vars_list_distinct: "distinct(fresh_vars_list n vs)"
  using Var_inj[of "fresh_strings_list ''x'' 0 vs n"]
    fresh_name_gen_for_strings_list[of "''x''" 0]
    distinct_map[of "Var" "fresh_strings_list ''x'' 0 vs n",symmetric]
  unfolding fresh_name_gen_list_def fresh_vars_list_def by auto

definition flat_ctxts :: "string set \<Rightarrow> ('f\<times>nat)set \<Rightarrow> ('f,string)ctxt set"
where "flat_ctxts tabus fs \<equiv> (\<Union>(f,n)\<in>fs.
  {flat_ctxt_i (map Var (fresh_vars n tabus)) f i | i. i \<in> {0..<n}}
)"

definition flat_ctxts_list :: "string list \<Rightarrow> ('f\<times>nat)list \<Rightarrow> ('f,string)ctxt list"
where "flat_ctxts_list tabus fs \<equiv> concat(map (\<lambda>(f,n). (
  let xs = map Var (fresh_vars_list n tabus) in
  map (flat_ctxt_i xs f) [(0::nat)..<n]
)) fs)"

lemma flat_ctxts_list_sound[simp]:
  assumes "distinct fs"
  shows "set(flat_ctxts_list tabus fs) = flat_ctxts (set tabus) (set fs)"
using assms by (induct fs) (auto simp: flat_ctxts_list_def flat_ctxts_def,force)

definition flat_ctxt_closure :: "('f\<times>nat)set \<Rightarrow> ('f,string)trs \<Rightarrow> ('f,string)trs" ("FC")
where "FC F R \<equiv> (\<Union>(l,r)\<in>(R)a. {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> flat_ctxts (vars_trs R) F}) \<union> (R - (R)a)"

definition flat_ctxt_closure_list :: "('f\<times>nat)list \<Rightarrow> ('f,string)trsL \<Rightarrow> ('f,string)trsL"
where "flat_ctxt_closure_list fs trs \<equiv> (
  let fcs      = flat_ctxts_list (vars_trs_list trs) fs in
  let (ras,rs) = partition (\<lambda>(l,r). root l \<noteq> root r) trs in
  flat_map (\<lambda>(l,r). map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) fcs) ras
  @ rs
)"

lemma flat_ctxts_empty[simp]: "flat_ctxts xs {} = {}" unfolding flat_ctxts_def by simp

lemma root_altering_list[simp]: "set(fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs)) = (set trs)a"
  unfolding partition_filter_conv root_altering_def by auto

lemma root_altering_list_aux[simp]: "set(snd(partition (\<lambda>(l,r). root l \<noteq> root r) trs)) = set trs - (set trs)a"
  unfolding partition_filter_conv root_altering_def by auto

lemma flat_ctxt_closure_list_sound_1[simp]: 
  fixes trs::"('f,string)trsL" and fs::"('f\<times>nat)list"
  assumes "distinct fs"
  shows "set(
  flat_map (\<lambda>(l,r).
    map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) (flat_ctxts_list (vars_trs_list trs) fs)
  ) (fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs))
) = (\<Union>(l,r)\<in>(set trs)a.
  {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> flat_ctxts (vars_trs(set trs)) (set fs)}
)"
proof -
  let ?C = "\<lambda>(l,r)C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)"
  let ?ra = "fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs)"
  let ?fcs = "flat_ctxts_list (vars_trs_list trs) fs"
  have "set(flat_map  (\<lambda>(l,r). map (?C(l,r)) ?fcs) ?ra) =
    (\<Union>x\<in>set(map (\<lambda>(l,r). map (?C(l,r)) ?fcs) ?ra). set x)" by simp
  also have "\<dots> = (\<Union>x\<in>(\<lambda>(l,r). map (?C(l,r)) ?fcs) ` set ?ra. set x)" by simp
  also have "\<dots> = (\<Union>(l,r)\<in>set ?ra. set((\<lambda>(l,r). map (?C(l,r)) ?fcs) (l,r)))" by best
  also have "\<dots> = (\<Union>(l,r)\<in>set ?ra. set(map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) ?fcs))" by simp
  also have "\<dots> = (\<Union>(l,r)\<in>set ?ra. {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> set(?fcs)})" by auto
  also have "\<dots> = (\<Union>(l,r)\<in>(set trs)a. {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> set(?fcs)})"
    using root_altering_list[of "trs"] by simp
  finally show ?thesis
  using flat_ctxts_list_sound[of "fs" "vars_trs_list trs",OF `distinct fs`] by auto
qed

lemma flat_ctxt_closure_list_sound[simp]:
  fixes trs::"('f,string)trsL"
  assumes "distinct fs"
  shows "set(flat_ctxt_closure_list fs trs) = FC(set fs)(set trs)"
proof -
  let ?fcs = "flat_ctxts_list (vars_trs_list trs) fs"
  let ?ras = "fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs)"
  let ?rs  = "snd(partition (\<lambda>(l,r). root l \<noteq> root r) trs)"
  have a: "set(
    let fcs      = ?fcs in
    let (ras,rs) = partition (\<lambda>(l,r). root l \<noteq> root r) trs in
    flat_map (\<lambda>(l,r). map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) fcs) ras @ rs) =
    set(flat_map (\<lambda>(l,r). map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) ?fcs) ?ras) \<union> set ?rs"
    by simp
  show ?thesis
    unfolding flat_ctxt_closure_list_def flat_ctxt_closure_def
    unfolding a
    unfolding flat_ctxt_closure_list_sound_1[OF `distinct fs`]
    unfolding root_altering_list_aux by simp
qed

lemma fccE[elim]:
  fixes R::"('f,string)trs"
  assumes "(s,t) \<in> FC F R"
  and ra: "(s,t) \<in> R - (R)a \<Longrightarrow> P"
  and rp: "\<exists>(l,r)\<in>(R)a.\<exists>C\<in>flat_ctxts (vars_trs R) F. s = C\<langle>l\<rangle> \<and> t = C\<langle>r\<rangle> \<Longrightarrow> P"
  shows "P"
using assms unfolding flat_ctxts_def flat_ctxt_closure_def by force

lemma fccI_altering[intro]:
  fixes R::"('f,string)trs" assumes "C \<in> flat_ctxts (vars_trs R) F"
  and "(l,r) \<in> (R)a"
  shows "(C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> FC F R"
using assms unfolding flat_ctxt_closure_def by auto

lemma fccI_preserving[intro]:
  fixes R::"('f,string)trs" assumes "(l,r) \<in> R - (R)a" shows "(l,r) \<in> FC F R"
using assms unfolding flat_ctxt_closure_def by simp

lemma partial_fcc_preserves_SN:
  fixes R::"('f,string)trs"
  assumes SN: "SN(rstep R)"
    and refl: "\<forall>(l,r)\<in>R'.\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    shows "SN(rstep R')"
using SN proof (rule contrapos_pp)
  let ?r = "rstep R"
  let ?c = "rstep R'"
  assume "\<not> SN ?c"
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?c" unfolding SN_defs by auto
  have "\<forall>i. (t i,t(Suc i)) \<in> ?r"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?c" by simp
    from rstep_imp_C_s_r[OF this]
      obtain C \<sigma> l r where fc: "(l,r) \<in> R'" and s: "t i = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t(Suc i) = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    from fc refl have "\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>" by auto
    then obtain u v C' where R: "(u,v) \<in> R" and l: "l = C'\<langle>u\<rangle>" and r: "r = C'\<langle>v\<rangle>" by auto
    from R have "(u\<cdot>\<sigma>,v\<cdot>\<sigma>) \<in> CS R" by auto
    hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?r" unfolding l r ctxt_sub_distr by auto
    thus "(t i,t(Suc i)) \<in> ?r" unfolding s t by (rule rstep_imp_ctxt_rstep)
  qed
  thus "\<not> SN ?r" unfolding SN_defs by auto
qed

(*
lemma partial_fcc_preserves_SN:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars R) F"
  assumes SN: "SN(rstep R)"
    and refl: "\<forall>(l,r)\<in>R'. (l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
    shows "SN(rstep R')"
using SN proof (rule contrapos_pp)
  let ?r = "rstep R"
  let ?c = "rstep R'"
  assume "\<not> SN ?c"
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?c" unfolding SN_defs by auto
  have "\<forall>i. (t i,t(Suc i)) \<in> ?r"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?c" by simp
    from rstep_imp_C_s_r[OF this]
      obtain C \<sigma> l r where fc: "(l,r) \<in> R'" and s: "t i = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t(Suc i) = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    from fc refl have "(l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)" by auto
    thus "(t i,t(Suc i)) \<in> ?r"
    proof
      assume "(l,r) \<in> R"
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R" by auto
      thus ?thesis unfolding s t by auto
    next
      assume "\<exists>(u,v)\<in>R.\<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
      then obtain u v C' where R: "(u,v) \<in> R" and C': "C' \<in> fcs" and l: "l = C'\<langle>u\<rangle>" and r: "r = C'\<langle>v\<rangle>" by auto
      from R have "(u\<cdot>\<sigma>,v\<cdot>\<sigma>) \<in> CS R" by auto
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?r" unfolding l r ctxt_sub_distr by auto
      thus ?thesis unfolding s t by (rule rstep_imp_ctxt_rstep)
    qed
  qed
  thus "\<not> SN ?r" unfolding SN_defs by auto
qed
*)

lemma fcc_imp_refl:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  shows "\<forall>(l,r)\<in>FC F R. (\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
proof -
  {
    fix l r assume "(l,r) \<in> FC F R"
    hence "\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    proof
      assume "(l,r) \<in> R - (R)a"
      hence "(l,r) \<in> R" by simp
      moreover have "l = \<box>\<langle>l\<rangle>" by simp
      moreover have "r = \<box>\<langle>r\<rangle>" by simp
      ultimately show ?thesis by best
    next
      assume "\<exists>(x,y)\<in>(R)a.\<exists>C\<in>flat_ctxts (vars_trs R) F. l = C\<langle>x\<rangle> \<and> r = C\<langle>y\<rangle>"
      moreover have "(R)a \<subseteq> R" unfolding root_altering_def by auto
      ultimately show ?thesis by auto
    qed
  }
  thus ?thesis by auto
qed

(*
lemma fcc_imp_refl:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars R) F"
  shows "\<forall>(l,r)\<in>FC F R. (l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
proof -
  {
    fix l r assume "(l,r) \<in> FC F R"
    hence "(l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
    proof
      assume "(l,r) \<in> R - (R)a" thus ?thesis by auto
    next
      assume "\<exists>(x,y)\<in>(R)a.\<exists>C\<in>flat_ctxts (vars R) F. l = C\<langle>x\<rangle> \<and> r = C\<langle>y\<rangle>"
      moreover have "(R)a \<subseteq> R" unfolding root_altering_def by auto
      ultimately show ?thesis unfolding fcs_def by auto
    qed
  }
  thus ?thesis by auto
qed
*)

lemma fcc_imp_pres:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars_trs R) F"
  shows "\<forall>(l,r)\<in>R. (l,r) \<in> FC F R \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> FC F R)"
proof (intro ballI2)
  fix l r assume "(l,r) \<in> R"
  hence "(l,r) \<in> (R)a \<or> (l,r) \<in> R - (R)a" by auto
  thus "(l,r) \<in> FC F R \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> FC F R)"
  proof
    assume Ra: "(l,r) \<in> (R)a"
    thus ?thesis unfolding fcs_def using fccI_altering[of _ R F l r] by auto
  next
    assume "(l,r) \<in> R - (R)a" thus ?thesis using fccI_preserving by best
  qed
qed

lemma fcc_preserves_SN:
  fixes R::"('f,string)trs" assumes "SN(rstep R)" shows "SN(rstep(FC F R))"
using partial_fcc_preserves_SN[OF assms,of "FC F R"] fcc_imp_refl[of F R] by simp

(*
lemma fcc_preserves_SN:
  fixes R::"('f,string)trs" assumes "SN(rstep R)" shows "SN(rstep(FC F R))"
using partial_fcc_preserves_SN[OF assms,of "FC F R" F] fcc_imp_refl[of F R] by simp
*)

lemma fresh_vars_sub:
  assumes "finite V" and "x \<in> set(fresh_vars n V)" shows "Var x\<cdot>(\<sigma>|_V) = Var x"
using fresh_vars_fresh[OF `finite V`] assms by (cases \<sigma>) auto

lemma flat_ctxt_sub_apply[simp]:
  assumes "finite V" and "C \<in> flat_ctxts V F" shows "C\<langle>t\<rangle>\<cdot>(\<sigma>|_V) = C\<langle>t\<cdot>(\<sigma>|_V)\<rangle>"
proof -
  from assms obtain f n i
    where i:"i<n" and fn: "(f,n)\<in>F" and C: "C = flat_ctxt_i (map Var (fresh_vars n V)) f i" (is "_ = flat_ctxt_i ?V f i")
    unfolding flat_ctxts_def by auto
  have fresh: "map (\<lambda>t. t\<cdot>(\<sigma>|_V)) ?V = ?V"
    using map_idI[of "?V" "\<lambda>t. t\<cdot>(\<sigma>|_V)"] fresh_vars_sub[OF `finite V`] by auto
  from C have C: "C = More f (take i ?V) \<box> (drop (Suc i) ?V)"
    unfolding flat_ctxt_i_def by simp
  show ?thesis unfolding C unfolding ctxt_apply.simps sub_apply.simps unfolding map_append
    unfolding take_map[symmetric]
    unfolding map.simps
    unfolding drop_map[symmetric]
    unfolding fresh by simp
qed

lemma flat_ctxt_has_fresh_vars:
  assumes finite: "finite V" and FC: "C \<in> flat_ctxts V F"
  shows "vars_ctxt C \<inter> V = {}"
proof -
  from FC obtain f n i where "(f,n) \<in> F" and "i < n" and "C = flat_ctxt_i (map Var (fresh_vars n V)) f i" (is "_ = flat_ctxt_i ?V _ _")
    unfolding flat_ctxts_def by auto
  hence C: "C = More f (take i ?V) \<box> (drop (Suc i) ?V)" unfolding flat_ctxt_i_def by simp
  have vC: "vars_ctxt C = (\<Union>set(map vars_term (take i ?V @ drop (Suc i) ?V)))" by (simp add: C)
  have"vars_ctxt C \<subseteq> \<Union>set(map vars_term ?V)" using fresh_vars_length[OF finite,of n] and `i < n` unfolding vC
    using set_take_subset[of i ?V] set_drop_subset[of "Suc i" ?V] by auto
  hence "{Var x | x. x \<in> vars_ctxt C} \<subseteq> set ?V" unfolding fresh_vars_def by auto
  with fresh_vars_fresh[OF finite,of n] show ?thesis by auto
qed

(* --> Position *)
lemma pos_exhaust_rev[case_names empty snoc]:
  assumes "p = \<epsilon> \<Longrightarrow> P" and "\<And>\<pi> i. p = \<pi>#>i \<Longrightarrow> P" shows "P"
using assms by (induct p rule:rev_induct) auto

lemma wf_FC: assumes "wf_trs R" shows "wf_trs(FC F R)"
unfolding wf_trs_def proof (intro allI impI)
  fix l r assume "(l,r) \<in> FC F R"
  thus "(\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
  proof
    assume "(l,r) \<in> R - (R)a" with assms show ?thesis unfolding wf_trs_def by auto
  next
    assume "\<exists>(l',r')\<in>(R)a. \<exists>C\<in>flat_ctxts (vars_trs R) F. l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    then obtain l' r' C where "(l',r') \<in> (R)a" and C: "C \<in> flat_ctxts (vars_trs R) F" and l: "l = C\<langle>l'\<rangle>" and r: "r = C\<langle>r'\<rangle>" by best
    hence R: "(l',r') \<in> R" unfolding root_altering_def by auto
    from C obtain f n i where "(f,n) \<in> F" and "i < n"
      and "C = flat_ctxt_i (map Var (fresh_vars n (vars_trs R))) f i" (is "C = flat_ctxt_i ?V f i") unfolding flat_ctxts_def by auto
    hence C: "C = More f (take i ?V) \<box> (drop (Suc i) ?V)" unfolding flat_ctxt_i_def by simp
    hence "l = Fun f (take i ?V@l'#drop (Suc i) ?V)" unfolding l by simp
    hence "\<exists>f ss. l = Fun f ss" by auto
    moreover have "vars_term r \<subseteq> vars_term l"
    proof -
      from R and assms have "vars_term r' \<subseteq> vars_term l'" unfolding wf_trs_def by auto
      with l r show ?thesis unfolding C by auto
    qed
    ultimately show ?thesis by simp
  qed
qed

(* --> Util *)
lemma distinct_take_drop: assumes dist: "distinct vs" and len: "i < length vs" shows "distinct(take i vs @ drop (Suc i) vs)" (is "distinct(?xs@?ys)")
proof -
  from id_take_nth_drop[OF len] have vs[symmetric]: "vs = ?xs @ vs!i # ?ys" .
  with dist have "distinct ?xs" and "distinct(vs!i#?ys)" and "set ?xs \<inter> set(vs!i#?ys) = {}" using distinct_append[of ?xs "vs!i#?ys"] by auto
  hence "distinct ?ys" and "set ?xs \<inter> set ?ys = {}" by auto
  with `distinct ?xs` show ?thesis using distinct_append[of ?xs ?ys] vs by simp
qed

(* --> Term *)
lemma sub_extend_flat_ctxt:
  assumes dist: "distinct vs"
    and len1: "length(take i (map Var vs)) = length ss1"
    and len2: "length(drop (Suc i) (map Var vs)) = length ss2"
    and i: "i < length vs"
  shows "More f (take i (map Var vs)) \<box> (drop (Suc i) (map Var vs)) \<cdot>c sub_extend \<sigma> (zip (take i vs@drop (Suc i) vs) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var vs"
  let ?vs1 = "take i vs" and ?vs2 = "drop (Suc i) vs"
  let ?ss1 = "take i ?V" and ?ss2 = "drop (Suc i) ?V"
  let ?\<sigma> = "sub_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" using i by simp
  from distinct_take_drop[OF dist i] have "distinct(?vs1@?vs2)" .
  from sub_extend_absorb[OF this len,of "\<sigma>"] have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding take_map drop_map map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" by auto
  ultimately show ?thesis by simp
qed

lemma distinct_map_Var: assumes "distinct xs" shows "distinct(map Var xs)"
using assms by (induct xs) auto

lemma fc_flat:
  assumes C: "C \<in> flat_ctxts V F"
    and finite: "finite V"
  shows "vars_ctxt C \<inter> V = {} \<and> (\<exists>f ss1 ss2. C = More f ss1 \<box> ss2
    \<and> (f,Suc(length(ss1@ss2))) \<in> F
    \<and> (\<forall>t\<in>set(ss1@ss2). is_var t)
    \<and> distinct(ss1@ss2))"
proof -
  from C obtain g i m where "i < m" and "(g,m) \<in> F"
    and C': "C = More g (take i (map Var (fresh_vars m V))) \<box> (drop (Suc i) (map Var (fresh_vars m V)))" (is "C = More g ?ss1 \<box> ?ss2")
    unfolding flat_ctxts_def flat_ctxt_i_def by auto
  from flat_ctxt_has_fresh_vars[OF finite C] have fresh: "vars_ctxt C \<inter> V = {}" .
  from fresh_vars_length[OF finite] have len: "length(fresh_vars m V) = m" by simp
  from fresh_vars_distinct[OF finite] have dist: "distinct(fresh_vars m V)" by simp
  from length_take[of i "fresh_vars m V"] `i < m` have take: "length(?ss1) = i" unfolding len by simp
  from length_drop[of "Suc i" "fresh_vars m V"] have drop: "length(?ss2) = length(fresh_vars m V) - Suc i" by simp
  from take drop have m: "Suc(length(?ss1@?ss2)) = m" unfolding len using `i < m` by simp
  have F:"(g,Suc(length(?ss1@?ss2))) \<in> F" using `(g,m) \<in> F` unfolding m .
  have "\<forall>t\<in>set ?ss1. is_var t" unfolding take_map by auto
  moreover have "\<forall>t\<in>set ?ss2. is_var t" unfolding drop_map by auto
  ultimately have is_var: "\<forall>t\<in>set (?ss1@?ss2). is_var t" by auto
  have dist: "distinct(?ss1@?ss2)" using distinct_take_drop[OF dist,unfolded len,OF `i < m`]
    unfolding take_map drop_map map_append[symmetric] by (rule distinct_map_Var)
  with fresh C' F is_var dist show ?thesis by auto
qed

(*
lemma partial_fcc_reflects_SN:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars R) F"
  assumes finite: "finite R"
    and subset: "funas R \<subseteq> F"
    and ne: "fcs \<noteq> {}"
    and pres: "\<forall>(l,r)\<in>R. (l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')"
    and SN: "SN(rstep R')"
  shows "SN(rstep R)"
using SN proof (rule contrapos_pp)
  let ?r = "wfrstep R"
  let ?c = "rstep R'"
  assume "\<not> SN(rstep R)"
  hence "\<not> SN ?r" unfolding SN_wfrstep_SN_rstep_conv .
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?r" unfolding SN_defs by auto
  from ne obtain C where C: "C \<in> fcs" by auto
  have "\<forall>i. (C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?r" by simp
    hence "(t i,t(Suc i)) \<in> rstep R \<and> t i \<in> Term(funas R) \<and> t(Suc i) \<in> Term(funas R)" unfolding wfrstep_def by fast
    hence rstep: "(t i,t(Suc i)) \<in> rstep R" and wfti: "t i \<in> Term(funas R)" and wftsi: "t(Suc i) \<in> Term(funas R)" by auto
    from rstep_imp_C_s_r[OF rstep] obtain D \<sigma> l r where R: "(l,r) \<in> R"
      and ti: "t i = D\<langle>l\<cdot>\<sigma>\<rangle>" and tsi: "t(Suc i) = D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    have vl: "vars l \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    have vr: "vars r \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    from R pres have "(l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')" by auto
    thus "(C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
    proof
      assume "(l,r) \<in> R'"
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R'" by auto
      hence "(D\<langle>l\<cdot>\<sigma>\<rangle>,D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c" by auto
      from rstep_imp_ctxt_rstep[OF this] show ?thesis unfolding ti tsi by simp
    next
      assume closed: "\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'"
      show ?thesis
      proof (cases D rule: ctxt_exhaust_rev)
        case Hole
        let ?\<sigma> = "\<sigma>|_(vars_trs R)"
        from C and closed have "(C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'" by auto
        from csI[OF this] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> CS R'" by simp
        from ccI[OF this,of \<box>] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> ?c" by simp
        hence "(C \<cdot>c ?\<sigma> \<langle>l\<cdot>?\<sigma>\<rangle>,C \<cdot>c ?\<sigma> \<langle>r\<cdot>?\<sigma>\<rangle>) \<in> ?c" by simp
        hence "(C \<cdot>c ?\<sigma> \<langle>l\<cdot>\<sigma>\<rangle>,C \<cdot>c ?\<sigma> \<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c"
          unfolding coincidence_lemma'[OF vl,symmetric] coincidence_lemma'[OF vr,symmetric] coincidence_lemma[symmetric] .
        thus ?thesis unfolding sub_apply_ctxt_id[OF flat_ctxt_has_fresh_vars,OF `finite R` C[unfolded fcs_def]]
          unfolding ti tsi Hole by simp
      next
        case (More E f ss1 ss2)
        from ctxt_apply_wf_imp_ctxt_and_term_wf[OF wfti[unfolded ti]]
          have D: "D \<in> Ctxt(funas R)" and "l\<cdot>\<sigma> \<in> Term(funas R)" by auto
        from ctxt_compose_wf_imp_ctxts_wf[OF D[unfolded More]] have "E \<in> Ctxt(funas R) \<and> More f ss1 \<box> ss2 \<in> Ctxt(funas R)" .
        hence "More f ss1 \<box> ss2 \<in> Ctxt(funas R)" by simp
        hence "(f,Suc(length(ss1@ss2))) \<in> funas R" (is "(f,?n) \<in> funas R") by (cases rule: wf_fun_ctxts.cases) auto
        with subset have F: "(f,?n) \<in> F" by auto
        let ?V = "fresh_vars ?n (vars R)"
        have len: "length ss1 < ?n" (is "?i < _") by simp
        with F have C': "More f (take ?i (map Var ?V)) \<box> (drop (Suc ?i) (map Var ?V)) \<in> flat_ctxts (vars R) F" (is "?C' \<in> flat_ctxts (vars R) F")
          unfolding flat_ctxts_def flat_ctxt_i_def by force
        with closed have FC_step: "(?C'\<langle>l\<rangle>,?C'\<langle>r\<rangle>) \<in> R'" unfolding fcs_def by best
        let ?\<sigma> = "sub_extend \<sigma> (zip (take ?i ?V @ drop (Suc ?i) ?V) (ss1@ss2))"
        from finite have "finite(vars R)" unfolding vars_trs_def vars_rule_def by simp
        from fresh_vars_fresh[OF this] have "vars R \<inter> set ?V = {}" by auto
        moreover have "set(take (length ss1) ?V) \<subseteq> set ?V" by (rule set_take_subset)
        moreover have "set(drop (Suc(length ss1)) ?V) \<subseteq> set ?V" by (rule set_drop_subset)
        ultimately have fresh: "vars R \<inter> set(take (length ss1) ?V @ drop (Suc(length ss1)) ?V) = {}" by auto
        from fresh_vars_distinct[OF `finite(vars R)`] have dist: "distinct ?V" .
        from len have len1: "length(take (length ss1) ?V) = length ss1" using fresh_vars_length[OF `finite(vars R)`,of ?n] by simp
        hence len2: "length(drop (Suc(length ss1)) ?V) = length ss2" using fresh_vars_length[OF `finite(vars R)`,of ?n] by auto
        from len1 have len1': "length(take (length ss1) (map Var ?V)) = length ss1" by simp
        from len2 have len2': "length(drop (Suc(length ss1)) (map Var ?V)) = length ss2" by simp
        have len3: "length ss1 < length ?V" using fresh_vars_length[OF `finite(vars R)`,of ?n] by simp
        have Cl: "(?C'\<langle>l\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>l\<cdot>\<sigma>\<rangle>"
          unfolding ctxt_sub_distr sub_extend_id[OF fresh vl,of \<sigma>] sub_extend_flat_ctxt[OF dist len1' len2' len3] by simp
        have Cr: "(?C'\<langle>r\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>r\<cdot>\<sigma>\<rangle>"
          unfolding ctxt_sub_distr sub_extend_id[OF fresh vr,of \<sigma>] sub_extend_flat_ctxt[OF dist len1' len2' len3] by simp
        let ?C'' = "More f ss1 \<box> ss2"
        from csI[OF FC_step] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> CS R'" unfolding Cl[symmetric] Cr[symmetric] by simp
        from ccI[OF this,of \<box>] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c" by simp
        from ccI[OF this,of E] have "(E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>,E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>) \<in> ?c" by simp
        from ccI[OF this,of C] have "(C\<langle>E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>,C\<langle>E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>) \<in> ?c" by simp
        thus ?thesis unfolding ti tsi More ctxt_compose_apply_eq .
      qed
    qed
  qed
  thus "\<not> SN ?c" unfolding SN_defs by auto
qed
*)

fun is_flat_ctxt :: "'v set \<Rightarrow> ('f\<times>nat)set \<Rightarrow> ('f,'v)ctxt \<Rightarrow> bool"
where "is_flat_ctxt V F (More f ss1 \<box> ss2) = (
          (f,Suc(length(ss1@ss2))) \<in> F
        \<and> forall is_var (ss1@ss2)
        \<and> distinct(ss1@ss2)
        \<and> vars_ctxt(More f ss1 \<box> ss2) \<inter> V = {}
      )"
    | "is_flat_ctxt V F _ = False"

lemma is_flat_ctxt_cases[consumes 1,case_names True]:
  assumes "is_flat_ctxt V F C"
    and "\<And>f ss1 ss2. C = More f ss1 \<box> ss2 \<Longrightarrow> P"
  shows "P"
using assms(1) proof (cases rule: is_flat_ctxt.cases[of "(V,F,C)"])
  case (1 V' F' f ss1 ss2)
  hence "C = More f ss1 \<box> ss2" by simp
  with assms(2) show ?thesis by auto
next
  case ("2_1" V' F') with assms(1) show ?thesis by simp
next
  case ("2_2" V' F' f ss1 g ts1 D ts2 ss2) with assms(1) show ?thesis by simp
qed

lemma is_flat_ctxt_imp_flat_ctxt: 
  assumes "is_flat_ctxt V F C"
  shows "vars_ctxt C \<inter> V = {} \<and> (\<exists>f ss1 ss2. C = More f ss1 \<box> ss2
          \<and> (f,Suc(length(ss1@ss2))) \<in> F
          \<and> forall is_var (ss1@ss2)
          \<and> distinct(ss1@ss2))"
using assms by (cases rule: is_flat_ctxt_cases,auto)

lemma terms_to_vars: assumes "forall is_var ts" shows "\<Union>set(map vars_term ts) = set(map the_var ts)"
using assms proof (induct ts)
  case Nil thus ?case by simp
next
  case (Cons t ts)
  hence IH: "\<Union>set(map vars_term ts) = set(map the_var ts)" by simp
  from Cons have "is_var t" by simp
  hence "vars_term t = {the_var t}" by (induct t) auto
  with IH show ?case by simp
qed

lemma Var_the_var_id: assumes "forall is_var ts" shows "map Var (map the_var ts) = ts"
using assms proof (induct ts)
  case Nil show ?case by simp
next
  case (Cons t ts)
  hence IH: "map Var (map the_var ts) = ts" and "is_var t" by auto
  from `is_var t` have "Var(the_var t) = t" by (induct t) auto
  with IH show ?case by simp
qed

lemma distinct_the_vars:
  assumes "forall is_var ts" and "distinct ts" shows "distinct(map the_var ts)"
using assms proof (induct ts)
  case Nil show ?case by simp
next
  case (Cons t ts)
  hence IH: "distinct(map the_var ts)"
    and "distinct ts" and "forall is_var ts" and "t \<notin> set ts" and "is_var t" by auto
  have "the_var t \<notin> set(map the_var ts)"
  proof
    assume "the_var t \<in> set(map the_var ts)"
    hence "Var(the_var t) \<in> set(map Var (map the_var ts))" unfolding map_map by auto
    moreover have "set(map Var (map the_var ts)) = set ts" using Var_the_var_id[OF `forall is_var ts`] by simp
    ultimately have "Var(the_var t) \<in> set ts" by best
    with `is_var t` have "t \<in> set ts" by (induct t) auto
    with `t \<notin> set ts` show False by simp
  qed
  with IH show ?case by simp
qed

lemma sub_extend_flat_ctxt':
  assumes dist: "distinct(vs1@vs2)"
    and len1: "length vs1 = length ss1"
    and len2: "length vs2 = length ss2"
  shows "More f (map Var vs1) \<box> (map Var vs2) \<cdot>c sub_extend \<sigma> (zip (vs1@vs2) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var (vs1@vs2)"
  let ?vs1 = "vs1" and ?vs2 = "vs2"
  let ?ss1 = "map Var vs1" and ?ss2 = "map Var vs2"
  let ?\<sigma> = "sub_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" by simp
  from sub_extend_absorb[OF dist len,of "\<sigma>"] have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" by auto
  ultimately show ?thesis by simp
qed

lemma partial_fcc_reflects_SN:
  fixes R::"('f,string)trs"
  assumes flat_ctxts: "\<forall>C\<in>fcs. is_flat_ctxt (vars_trs R) F C"
    and complete: "\<forall>(f,n)\<in>F.\<forall>i<n.\<exists>ss1 ss2. (More f ss1 \<box> ss2) \<in> fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    and subset: "funas_trs R \<subseteq> F"
    and ne: "fcs \<noteq> {}"
    and pres: "\<forall>(l,r)\<in>R. (l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')"
    and SN: "SN(rstep R')"
  shows "SN(rstep R)"
using SN proof (rule contrapos_pp)
  let ?r = "wfrstep R"
  let ?c = "rstep R'"
  assume "\<not> SN(rstep R)"
  hence "\<not> SN ?r" unfolding SN_wfrstep_SN_rstep_conv .
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?r" unfolding SN_defs by auto
  from ne obtain C where C: "C \<in> fcs" by auto
  from flat_ctxts[THEN bspec[where x=C],OF C] have distinct: "vars_ctxt C \<inter> vars_trs R = {}"
    using is_flat_ctxt_imp_flat_ctxt by best
  have "\<forall>i. (C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?r" by simp
    hence "(t i,t(Suc i)) \<in> rstep R \<and> t i \<in> Term(funas_trs R) \<and> t(Suc i) \<in> Term(funas_trs R)" unfolding wfrstep_def by fast
    hence rstep: "(t i,t(Suc i)) \<in> rstep R" and wfti: "t i \<in> Term(funas_trs R)" and wftsi: "t(Suc i) \<in> Term(funas_trs R)" by auto
    from rstep_imp_C_s_r[OF rstep] obtain D \<sigma> l r where R: "(l,r) \<in> R"
      and ti: "t i = D\<langle>l\<cdot>\<sigma>\<rangle>" and tsi: "t(Suc i) = D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    have vl: "vars_term l \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    have vr: "vars_term r \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    from R pres have "(l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')" by auto
    thus "(C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
    proof
      assume "(l,r) \<in> R'"
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R'" by auto
      hence "(D\<langle>l\<cdot>\<sigma>\<rangle>,D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c" by auto
      from rstep_imp_ctxt_rstep[OF this] show ?thesis unfolding ti tsi by simp
    next
      assume closed: "\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'"
      show ?thesis
      proof (cases D rule: ctxt_exhaust_rev)
        case Hole
        let ?\<sigma> = "\<sigma>|_(vars_trs R)"
        from C and closed have "(C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'" by auto
        from csI[OF this] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> CS R'" by simp
        from ccI[OF this,of \<box>] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> ?c" by simp
        hence "(C \<cdot>c ?\<sigma> \<langle>l\<cdot>?\<sigma>\<rangle>,C \<cdot>c ?\<sigma> \<langle>r\<cdot>?\<sigma>\<rangle>) \<in> ?c" by simp
        hence "(C \<cdot>c ?\<sigma> \<langle>l\<cdot>\<sigma>\<rangle>,C \<cdot>c ?\<sigma> \<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c"
          unfolding coincidence_lemma'[OF vl,symmetric] coincidence_lemma'[OF vr,symmetric] coincidence_lemma[symmetric] .
        thus ?thesis unfolding sub_apply_ctxt_id[OF distinct] unfolding ti tsi Hole by simp
      next
        case (More E f ss1 ss2)
        from ctxt_apply_wf_imp_ctxt_and_term_wf[OF wfti[unfolded ti]]
          have D: "D \<in> Ctxt(funas_trs R)" and "l\<cdot>\<sigma> \<in> Term(funas_trs R)" by auto
        from ctxt_compose_wf_imp_ctxts_wf[OF D[unfolded More]] have "E \<in> Ctxt(funas_trs R) \<and> More f ss1 \<box> ss2 \<in> Ctxt(funas_trs R)" .
        hence "More f ss1 \<box> ss2 \<in> Ctxt(funas_trs R)" by simp
        hence "(f,Suc(length(ss1@ss2))) \<in> funas_trs R" (is "(f,?n) \<in> funas_trs R") by (cases rule: wf_fun_ctxts.cases) auto
        with subset have F: "(f,?n) \<in> F" by auto
        have len: "length ss1 < ?n" (is "?i < _") by simp
        from complete[THEN bspec[where x="(f,?n)"],OF F] len
          obtain  ts1 ts2 where C': "More f ts1 \<box> ts2 \<in> fcs" (is "?C' \<in> fcs")
          and "length ts1 = ?i" and "length ts2 = ?n - ?i - 1" by best
        with closed have FC_step: "(?C'\<langle>l\<rangle>,?C'\<langle>r\<rangle>) \<in> R'" by best
        have len_ss1: "length ss1 = length ts1" using `length ts1 = ?i` by simp
        have len_ss2: "length ss2 = length ts2" using `length ts2 = ?n - ?i - 1` by simp
        have len_ss1ss2: "length(ss1@ss2) = length(ts1@ts2)" using len_ss1 len_ss2 by simp
        let ?V = "map the_var (ts1@ts2)"
        let ?\<sigma> = "sub_extend \<sigma> (zip ?V (ss1@ss2))"
        from is_flat_ctxt_imp_flat_ctxt[OF flat_ctxts[THEN bspec,OF C']]
          have is_var: "forall is_var (ts1@ts2)" and fresh: "vars_ctxt ?C' \<inter> vars_trs R = {}" by auto
        have vars_C': "vars_ctxt ?C' = set ?V" using terms_to_vars[OF is_var] by simp
        from fresh have fresh: "vars_trs R \<inter> set ?V = {}" unfolding vars_C' by auto
        from is_flat_ctxt_imp_flat_ctxt[OF flat_ctxts[THEN bspec,OF C']]
          have "forall is_var (ts1@ts2)" and "distinct(ts1@ts2)" by auto
        hence "distinct ?V" by (rule distinct_the_vars)
        hence dist: "distinct(map the_var ts1 @ map the_var ts2)" by simp
        have len1: "length(map the_var ts1) = length ss1" unfolding len_ss1 by simp
        have len2: "length(map the_var ts2) = length ss2" unfolding len_ss2 by simp
        from is_var have "forall is_var ts1" by simp
        hence id1: "map Var (map the_var ts1) = ts1" by (rule Var_the_var_id)
        from is_var have "forall is_var ts2" by simp
        hence id2: "map Var (map the_var ts2) = ts2" by (rule Var_the_var_id)
        have Cl: "(?C'\<langle>l\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>l\<cdot>\<sigma>\<rangle>"
          unfolding ctxt_sub_distr sub_extend_id[OF fresh vl,of \<sigma>]
          unfolding map_append sub_extend_flat_ctxt'[OF dist len1 len2,unfolded id1 id2] by simp
        have Cr: "(?C'\<langle>r\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>r\<cdot>\<sigma>\<rangle>"
          unfolding ctxt_sub_distr sub_extend_id[OF fresh vr,of \<sigma>]
          unfolding map_append sub_extend_flat_ctxt'[OF dist len1 len2,unfolded id1 id2] by simp
        let ?C'' = "More f ss1 \<box> ss2"
        from csI[OF FC_step] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> CS R'" unfolding Cl[symmetric] Cr[symmetric] by simp
        from ccI[OF this,of \<box>] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c" by simp
        from ccI[OF this,of E] have "(E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>,E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>) \<in> ?c" by simp
        from ccI[OF this,of C] have "(C\<langle>E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>,C\<langle>E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>) \<in> ?c" by simp
        thus ?thesis unfolding ti tsi More ctxt_compose_apply_eq .
      qed
    qed
  qed
  thus "\<not> SN ?c" unfolding SN_defs by auto
qed

lemma flat_ctxt_is_flat: assumes fin: "finite V" and fc: "C \<in> flat_ctxts V F" shows "is_flat_ctxt V F C"
using fc_flat[OF fc fin] by auto

lemma fcc_reflects_SN:
  fixes R::"('f,string)trs"
  assumes finite: "finite R"
    and subset: "funas_trs R \<subseteq> G"
    and ne: "flat_ctxts (vars_trs R) G \<noteq> {}" (is "?fcs \<noteq> {}")
    and SN: "SN(rstep(FC G R))"
  shows "SN(rstep R)"
proof -
  from finite have fin: "finite(vars_trs R)" unfolding vars_trs_def vars_rule_def by auto
  have fcs: "\<forall>C\<in>?fcs. is_flat_ctxt (vars_trs R) G C" using flat_ctxt_is_flat[OF fin] by best
  have complete: "\<forall>(f,n)\<in>G.\<forall>i<n.\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> ?fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
  proof (intro ballI2 allI impI)
    fix f n i
    let ?V = "map Var (fresh_vars n (vars_trs R))"
    assume "(f,n) \<in> G" and "i < n"
    hence "flat_ctxt_i ?V f i \<in> ?fcs" unfolding flat_ctxts_def by auto
    hence "More f (take i ?V) \<box> (drop (Suc i) ?V) \<in> ?fcs" (is "More f ?ss1 \<box> ?ss2 \<in> _") unfolding flat_ctxt_i_def by auto
    moreover have "length ?ss1 = i" using fresh_vars_length[OF fin,of n] `i < n` by simp
    moreover have "length ?ss2 = n - i - 1" using fresh_vars_length[OF fin,of n] `i < n` by simp
    ultimately show "\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> ?fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1" by best
  qed
  from partial_fcc_reflects_SN[OF fcs complete subset ne fcc_imp_pres SN] show ?thesis .
qed

lemma SN_iff_fcc_SN:
  fixes R::"('f,string)trs"
  assumes "finite R" and "funas_trs R \<subseteq> G" and "flat_ctxts (vars_trs R) G \<noteq> {}"
  shows "(SN(rstep R)::bool) = SN(rstep(FC G R))"
using fcc_reflects_SN[OF assms] fcc_preserves_SN[of R G] by best

lemma fcc_ne_imp_funs_ne:
  fixes R::"('f,string)trs"
  assumes ne: "flat_ctxts V (funas_trs R) \<noteq> {}"
  shows "funs_trs R \<noteq> {}"
proof -
  from ne have "funas_trs R \<noteq> {}" unfolding flat_ctxts_def by auto
  then obtain l r where R: "(l,r) \<in> R" and "funas_term l \<union> funas_term r \<noteq> {}" unfolding funas_rule_def funas_trs_def by force
  hence "funas_term l \<noteq> {} \<or> funas_term r \<noteq> {}" by simp
  thus "funs_trs R \<noteq> {}"
  proof
    assume "funas_term l \<noteq> {}"
    then obtain f ss where "l = Fun f ss" by (induct l) auto
    hence "f \<in> funs_term l" by simp
    with R show ?thesis unfolding funs_trs_def funs_rule_def by force
  next
    assume "funas_term r \<noteq> {}"
    then obtain f ss where "r = Fun f ss" by (induct r) auto
    hence "f \<in> funs_term r" by simp
    with R show ?thesis unfolding funs_trs_def funs_rule_def by force
  qed
qed

(* Closure under Flat Contexts *)
definition check_rule_preserving :: "('f::shows,string)ctxt list
  \<Rightarrow> ('f,string)rule list
  \<Rightarrow> ('f,string)rule
  \<Rightarrow> shows monad"
where "check_rule_preserving fcs rs' rule \<equiv> (
  checkB (rule mem rs' \<or> forall (\<lambda>C. (C\<langle>fst rule\<rangle>,C\<langle>snd rule\<rangle>) mem rs') fcs)
    (''the rule '' +#+ shows_rule rule +@+ '' is neither contained in the resulting set of rules nor closed under all flat contexts'' +#+ shows_nl)
)"

definition check_rule_reflecting :: "('f::shows,string)ctxt list
  \<Rightarrow> ('f,string)rule list
  \<Rightarrow> ('f,string)rule
  \<Rightarrow> shows monad"
where "check_rule_reflecting fcs rs rule \<equiv> (
  checkB (any (\<lambda>(l,r). any (\<lambda>C. fst rule = C\<langle>l\<rangle> \<and> snd rule = C\<langle>r\<rangle>) (\<box>#fcs)) rs)
    (''the rule '' +#+ shows_rule rule +@+ '' is neither contained in the original set of rules nor obtained by applying a flat context'' +#+ shows_nl)
)"

lemma check_rule_preserving_sound:
  assumes "isOK(check_rule_preserving fcs rs' (l,r))"
  shows "(l,r) \<in> set rs' \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> set rs')"
proof -
  from assms have "(l,r) mem rs' \<or> forall (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) mem rs') fcs" unfolding check_rule_preserving_def by simp
  thus ?thesis unfolding mem_iff by simp
qed

lemma check_rule_reflecting_sound:
  assumes "isOK(check_rule_reflecting fcs rs (l,r))"
  shows "\<exists>(u,v) \<in> set rs.\<exists>C\<in>set(\<box>#fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
proof -
  from assms have "any (\<lambda>(u,v). any (\<lambda>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>) (\<box>#fcs)) rs" unfolding check_rule_reflecting_def by simp
  thus ?thesis unfolding any_set_conv .
qed

lemma check_rules_preserving_sound:
  assumes "isOK(checkAllM (check_rule_preserving fcs lR) R)"
  shows "\<forall>(l,r)\<in>set R. (l,r) \<in> set lR \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> set lR)"
proof -
  from assms have "\<forall>rule\<in>set R. isOK(check_rule_preserving fcs lR rule)" by simp
  thus ?thesis using check_rule_preserving_sound by best
qed

lemma check_rules_reflecting_sound:
  assumes "isOK(checkAllM (check_rule_reflecting fcs R) lR)"
  shows "\<forall>(l,r)\<in>set lR. \<exists>(u,v)\<in>set R. \<exists>C\<in>set(\<box>#fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
proof -
  from assms have "\<forall>rule\<in>set lR. isOK(check_rule_reflecting fcs R rule)" by simp
  thus ?thesis using check_rule_reflecting_sound[of fcs R] by fast
qed

(* --> Term *)
instantiation ctxt :: ("shows","shows")"shows" begin
  fun shows_ctxt :: "('f::shows,'v::shows)ctxt \<Rightarrow> shows"
  where "shows_ctxt(Hole) = shows_string ''[]''"
      | "shows_ctxt(More f ss1 D ss2) = (
        shows f +@+ ''('' +#+
        shows_list_gen shows [] [] '','' '','' ss1 +@+
        shows_ctxt D +@+
        shows_list_gen shows '')'' '','' '','' '')'' ss2
      )"
  definition shows_list_ctxt where "shows_list_ctxt(cs::('f::shows,'v::shows)ctxt list) \<equiv> shows_list_aux shows cs"

  lemma assoc_ctxt: "shows (c::('f::shows,'v::shows)ctxt) r @ s = shows c (r@s)" by (induct c arbitrary: r s) auto
  lemma assoc_ctxt_elt: "\<forall>(c::('f::shows,'v::shows)ctxt)\<in>set cs. shows c r @ s = shows c (r@s)"
    using assoc_ctxt by auto
  instance by
    (intro_classes,rule assoc_ctxt,unfold shows_list_ctxt_def)
    (rule shows_list_aux_assoc[OF assoc_ctxt_elt])
end

definition check_flat_ctxt :: "string list \<Rightarrow> ('f::shows,string)ctxt \<Rightarrow> shows monad"
where "check_flat_ctxt vs C \<equiv> case C of
  More f ss1 \<box> ss2 \<Rightarrow> checkB (distinct(ss1@ss2)) (shows C +@+ '' contains duplicate variables'' +#+ shows_nl)
                   \<guillemotright> checkB (forall is_var (ss1@ss2)) (shows C +@+ '' is not flat, i.e., has depth greater than one'' +#+ shows_nl)
                   \<guillemotright> checkB (forall (\<lambda>t. \<not> (the_var t) mem vs) (ss1@ss2)) (shows C +@+ '' has to contain only fresh variables'' +#+ shows_nl)
| _                \<Rightarrow> error (shows C +@+ '' is not a flat context'' +#+ shows_nl)
"

fun is_flat_ctxt_list :: "'v list \<Rightarrow> ('f\<times>nat)list \<Rightarrow> ('f,'v)ctxt \<Rightarrow> bool"
where "is_flat_ctxt_list vs fas (More f ss1 \<box> ss2) = (
  (f,Suc(length(ss1@ss2))) mem fas
\<and> forall is_var (ss1@ss2) \<and> distinct(ss1@ss2) \<and> list_inter (map the_var (ss1@ss2)) vs = []
)"
    | "is_flat_ctxt_list vs fas _ = False"

lemma is_flat_ctxt_list_cases[consumes 1,case_names True]:
  assumes "is_flat_ctxt_list vs fas C" and "\<And>f ss1 ss2. C = More f ss1 \<box> ss2 \<Longrightarrow> P" shows "P"
using assms by (cases rule: is_flat_ctxt_list.cases[of "(vs,fas,C)"]) auto

lemma is_flat_ctxt_list_sound: "is_flat_ctxt_list vs fas C = is_flat_ctxt (set vs) (set fas) C"
proof
  assume fc: "is_flat_ctxt_list vs fas C" thus "is_flat_ctxt (set vs) (set fas) C"
  proof (cases rule: is_flat_ctxt_list_cases)
    case (True f ss1 ss2)
    from fc have "(f,Suc(length(ss1@ss2))) \<in> set fas"
      and "forall is_var (ss1@ss2)" and "distinct(ss1@ss2)"
      and "set(map the_var (ss1@ss2)) \<inter> set vs = {}"
      unfolding True  mem_iff[symmetric] list_inter_conv[symmetric] by auto
    moreover have "set(map the_var (ss1@ss2)) = vars_ctxt C" unfolding True
      using terms_to_vars[OF `forall is_var (ss1@ss2)`] by simp
    ultimately show ?thesis unfolding True by simp
  qed
next
  assume fc: "is_flat_ctxt (set vs) (set fas) C" thus "is_flat_ctxt_list vs fas C"
  proof (cases rule: is_flat_ctxt_cases)
    case (True f ss1 ss2)
       from fc have "(f,Suc(length(ss1@ss2))) mem fas"
      and "forall is_var (ss1@ss2)" and "distinct(ss1@ss2)"
      and "list_inter (map the_var (ss1@ss2)) vs = []"
      unfolding True mem_iff using terms_to_vars[of "ss1@ss2"] list_inter_conv[of "map the_var (ss1@ss2)" "vs"] by auto
    thus ?thesis unfolding True by simp
  qed
qed

fun hole_at where "hole_at n i f (More g ss1 \<box> ss2) = (g = f \<and> length ss1 = i \<and> length ss2 = n - i - 1)"
                | "hole_at n i f _                  = False"

definition check_flat_ctxt_complete :: "('f::shows,string)ctxt list \<Rightarrow> ('f\<times>nat) \<Rightarrow> shows monad"
where "check_flat_ctxt_complete fcs fa \<equiv>
  checkB (forall (\<lambda>i. any (hole_at (snd fa) i (fst fa)) fcs) [0..<snd fa]) (''the list of flat contexts is incomplete'' +#+ shows_nl)
"

definition check_is_flat_ctxt :: "string list \<Rightarrow> ('f::shows\<times>nat)list \<Rightarrow> ('f,string)ctxt \<Rightarrow> shows monad"
where "check_is_flat_ctxt vs fas C \<equiv> (
  checkB (is_flat_ctxt_list vs fas C) (shows C +@+ '' is not a flat context'' +#+ shows_nl)
)"

definition fcc_tt :: "('f::shows,string)trsL
  \<Rightarrow> ('f,string)ctxt list
  \<Rightarrow> ('f,string)trsL
  \<Rightarrow> shows monad"
where "fcc_tt R fcs lR \<equiv> (
  let vs  = vars_trs_list R in
  let fas = funas_trs_list R in (
    checkB (fcs \<noteq> []) (''at least one flat context is required for the closure under flat contexts'' +#+ shows_nl)
  \<guillemotright> checkAllM (check_flat_ctxt vs) fcs
  \<guillemotright> checkAllM (check_is_flat_ctxt vs fas) fcs
  \<guillemotright> checkAllM (check_flat_ctxt_complete fcs) fas
  \<guillemotright> checkAllM (check_rule_preserving fcs lR) R
  \<guillemotright> checkAllM (check_rule_reflecting fcs R) lR
  )
)"

lemma fcc_tt_sound: "sound_tt_impl(\<lambda>R rR. isOK(fcc_tt R fcs rR))"
unfolding sound_tt_impl_def
proof (intro allI impI)
  fix R rR assume ok: "isOK(fcc_tt R fcs rR)" and SN: "SN(rstep(set rR))"
  from ok[unfolded fcc_tt_def Let_def] have "\<forall>C\<in>set fcs. is_flat_ctxt (vars_trs(set R)) (funas_trs(set R)) C"
    using is_flat_ctxt_list_sound[of "vars_trs_list R" "funas_trs_list R"]
    check_is_flat_ctxt_def[of "vars_trs_list R" "funas_trs_list R"] by auto
  moreover have "funas(set R) \<subseteq> funas(set R)" by simp
  moreover
    have "\<forall>(f,n)\<in>funas_trs(set R).\<forall>i<n.\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
  proof (intro ballI2 allI impI)
    fix f n i assume fn: "(f,n) \<in> funas_trs(set R)" and i: "i < n"
    from ok[unfolded fcc_tt_def Let_def] fn i
      have "any (hole_at n i f) fcs" using check_flat_ctxt_complete_def[of fcs "(f,n)"] by auto
    then obtain C where fcs: "C\<in>set fcs" and C: "hole_at n i f C" unfolding any_set_conv by auto
    from C show "\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    proof (cases rule: hole_at.cases[of "(n,i,f,C)"])
      case (1 n' i' f' g ss1 ss2) with fcs C show ?thesis by auto
    next
      case ("2_1" n' i' f') with C show ?thesis by simp
    next
      case ("2_2" n' i' f' g ss1 h ts1 D ts2 ss2) with C show ?thesis by simp
    qed
  qed
  moreover from ok[unfolded fcc_tt_def Let_def] have "fcs \<noteq> []" by simp
  moreover from ok[unfolded fcc_tt_def Let_def] check_rules_preserving_sound[of fcs rR R]
  have "\<forall>(l,r)\<in>set R. (l,r) \<in> set rR \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> set rR)" by auto
  ultimately show "SN(rstep(set R))" using SN partial_fcc_reflects_SN[of "set fcs" "set R" "funas_trs(set R)" "set rR"] by simp
qed

lemma fcc_tt_sound_lab: "sound_lab_tt_impl(\<lambda>R m lR rm. rm = 0 \<and> isOK(fcc_tt R fcs lR))"
  using sound_tt_impl_to_sound_lab_tt_impl[OF fcc_tt_sound] .

locale rl_interpr =
  fixes F::"('f::shows,'l::shows)lab set"
    and f::"('f,'l)lab"
  assumes ne: "f \<in> F"
begin
fun rl_inter
where "rl_inter g _ = (if g \<in> F then g else f)"
end

locale rl_interpr_impl =
  fixes fs::"('f::shows,'l::shows)lab list"
    and f::"('f,'l)lab"
  assumes ne: "f mem fs"

sublocale rl_interpr_impl \<subseteq> rl_interpr "set fs"
proof
  from ne show "f \<in> set fs" unfolding mem_iff .
qed

fun rl_inter_list
where "rl_inter_list fs d g _  = (if g mem fs then g else d)"

context rl_interpr_impl
begin
lemma ne2: "f \<in> set fs" using ne unfolding mem_iff by simp

lemma rl_inter_list_sound[simp]:
  shows  "rl_inter_list fs f = rl_inter"
proof (intro ext)
  fix g cs show "rl_inter_list fs f g cs = rl_inter g cs" by (cases "g mem fs") (auto simp: mem_iff)
qed
end

sublocale rl_interpr \<subseteq> sl_interpr "F" "f" "rl_inter" "{}" L
proof (unfold_locales,rule ne)
  show "SN(rstep(decr_of_ord{}))" unfolding decr_of_ord_def by simp
  show "wf_inter rl_inter F" unfolding wf_inter_def
  proof (intro allI impI)
    fix f cs assume "set cs \<subseteq> F"
    show "rl_inter f cs \<in> F" using ne by auto
  qed
qed

context rl_interpr_impl
begin

lemma sl_interpr_impl: "sl_interpr_impl  (set fs) f (rl_inter_list fs f) {} L
  (check_sl_lab_trs (rl_inter_list fs f) L fs)
  (check_sl_model_lab_trs (rl_inter_list fs f) L fs)"
by (unfold_locales,rule ne2,simp,rule wf_I,rule Decr_SN,
  simp only: check_sl_lab_trs_sound,
  simp only: check_sl_model_lab_trs_sound,auto)
end

definition rl_fun where "rl_fun _ cs = Right cs"

definition rl_tt :: "(('f::shows,'l::shows)lab,string)trsL
  \<Rightarrow> (('f,'l)lab,string)trsL \<Rightarrow> shows monad"
where "rl_tt R lR \<equiv> (let fs = funs_trs_list R in (
   checkB (fs \<noteq> []) (''root-labeling requires at least one function symbol in the signature'' +#+ shows_nl)
 \<guillemotright> (let d = hd fs in (check_sem_lab_tt
     (check_sl_lab_trs (rl_inter_list fs d) rl_fun fs)
     (check_sl_model_lab_trs (rl_inter_list fs d) rl_fun fs)
     {}
     (rl_inter_list fs d) rl_fun R lR
))))"

(* TODO: move to Util *)
fun list_diff :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "list_diff xs [] = xs"
    | "list_diff xs (y#ys) = list_diff (removeAll y xs) ys"

(* The Root-Labeling Processor *)
definition rl_proc :: "(('f::shows,'l::shows)lab,string)trsL
  \<Rightarrow> (('f,'l)lab,string)trsL \<Rightarrow> (('f,'l)lab,string)trsL \<Rightarrow> (('f,'l)lab,string)trsL
  \<Rightarrow> shows monad"
where "rl_proc P R lP lR \<equiv> (let fs = list_union (funs_trs_list R) (list_diff (funs_trs_list P) (defs_list P)) in (
  checkB (fs \<noteq> []) (''root-labeling requires at least one function symbol in the signature'' +#+ shows_nl)
\<guillemotright> (let d = hd fs in (check_sem_lab_proc
    (check_sl_lab_trs (rl_inter_list fs d) rl_fun fs)
    (check_sl_model_lab_trs (rl_inter_list fs d) rl_fun fs)
    {}
    (rl_inter_list fs d) rl_fun P R lP lR
))))"

lemma rl_tt_sound: "sound_lab_tt_impl(\<lambda>R m lR rm. rm = Suc m \<and> isOK(rl_tt R lR))"
unfolding sound_lab_tt_impl_def
proof (intro allI impI)
  fix R::"(('f::shows,'l::shows)lab,string)trsL" and m rR rm
  assume check: "rm = Suc m \<and> isOK(rl_tt R rR)" and decr: "decrSN (set rR) rm"
  let ?fs = "funs_trs_list R"
  let ?def = "hd ?fs"
  from check have "?fs \<noteq> []" unfolding rl_tt_def by auto
  hence "?def mem ?fs" by (cases "?fs") simp_all
  hence rl: "rl_interpr_impl ?fs ?def" unfolding rl_interpr_impl_def rl_interpr_def .
  have rm: "rm = Suc m" using check by simp
  show "decrSN (set R) m"
    using sl_interpr_impl.check_sem_lab_tt_sound[OF rl_interpr_impl.sl_interpr_impl[OF rl,of rl_fun]] check
    unfolding sound_lab_tt_impl_def rl_tt_def
    using rm decr by (auto simp: Let_def)
qed

lemma rl_proc_sound: "sound_lab_proc_impl(\<lambda>P R m ruleMap lP lR rm. rm = Suc m \<and> isOK(rl_proc P R lP lR))"
unfolding sound_lab_proc_impl_def
proof (intro allI impI)
  fix P R::"(('f::shows,'l::shows)lab,string)trsL" and m lP lR rm
  assume check: "rm = Suc m \<and> isOK(rl_proc P R lP lR)" and finite: "finite_lab_dpp(set lP,set lR) rm"
  let ?fs = "list_union (funs_trs_list R) (list_diff (funs_trs_list P) (defs_list P))"
  let ?def = "hd ?fs"
  from check have "?fs \<noteq> []" unfolding rl_proc_def by auto
  hence ne: "?def mem ?fs" by (cases "?fs") simp_all
  hence rl: "rl_interpr_impl ?fs ?def" unfolding rl_interpr_impl_def rl_interpr_impl_def .
  have rm: "rm = Suc m" using check by simp
  show "finite_lab_dpp(set P,set R) m"
    using sl_interpr_impl.check_sem_lab_proc_sound[OF rl_interpr_impl.sl_interpr_impl[OF rl,of rl_fun]] check
    unfolding sound_lab_proc_impl_def rl_proc_def Let_def
    using rm finite by force
qed

end
